#include <u.h>
#include <libc.h>
#include <ctype.h>
#include "html.h"

typedef struct Keyword Keyword;

struct Keyword
{
	Rune *name;
	int	flag1;
	int	flag2;
};

enum
{
	NOTAG,
	TAG,
	TEXT,
};

enum
{
	EOF = -1,
	EOB = -2,
};

#include "data.c"



static int installed = 0;



static
void 
dprint(char c, HtmlParser *p, char *m, ...)
{
	static char buf[1024];
	char *out;
	va_list args;

	if(htmldebug[c]){
		va_start(args, m);
		out = buf+sprint(buf, "%S:%d: ", p->name, p->lineno);
		out += vsnprint(out, sizeof(buf)-(out-buf)-1, m, args);
		va_end(args);
		*out = '\0';
		fprint(2, "%s\n", buf);
	}
}

static
int 
binsearch(Rune *s, Keyword *kp, int n)
{
	int cond, l, m, h;

	l = 0;
	h = n - 1;
	while(l <= h){
		m = (l + h) / 2;
		if ((cond = runestrcmp(s, kp[m].name)) < 0)
			h = m - 1;
		else if (cond > 0)
			l = m + 1;
		else
			return m;
	}
	return -1;
}

static
int
runecistrncmp(Rune *s1, Rune *s2, int n)
{
	int c1, c2;

	while(*s1 && n-- > 0){
		c1 = *s1++;
		c2 = *s2++;

		if(c1 == c2)
			continue;

		if(c1 >= 'A' && c1 <= 'Z')
			c1 -= 'A' - 'a';

		if(c2 >= 'A' && c2 <= 'Z')
			c2 -= 'A' - 'a';

		if(c1 != c2)
			return c1 - c2;
	}
	if(n <= 0)
		return 0;
	return -*s2;
}


static
int
getc(HtmlParser *p)
{
	int c;

	if(p->yysptr > p->yysbuf)
		c  = (uchar)*--p->yysptr;
	else{
		if(--p->nin < 0){
			if(p->eof)
				c = EOF;
			else
				c =  EOB;
		}else
			c = *p->in++ & 0xff;
	}
	return c;
}

static
void
ungetstr(HtmlParser *p, char *s, int n)
{
	int i;

	for(i=n-1; i>=0; i--){
		if(p->yysptr >= p->yysbuf+sizeof(p->yysbuf))
			sysfatal("pushed back too much: %.20s...", p->yysbuf);
		*p->yysptr++ = s[i];
	}
}


static
int
input(HtmlParser *p)
{
	char s[UTFmax+1];
	int c, i;
	Rune r;

	c = getc(p);
	if(c >= Runeself){
		i = 0;
		s[i++] = c;
loop:
		c = getc(p);
		if(c==EOF || c==EOB)
			ungetstr(p, s, i);
		s[i++] = c;
		if(!fullrune(s, i))
			goto loop;

		s[i] = 0;
		chartorune(&r, s);
		c = r;
	}
	if(c == '\n')
		p->lineno++;

	return c;
}

static
void
unput(HtmlParser *p, Rune r)
{
	char s[UTFmax+1];
	int n;

	if(r == L'\n')
		p->lineno--;

	n = runetochar(s, &r);
	ungetstr(p, s, n);
}


static
void
unputstr(HtmlParser *p, Rune *s, int n)
{
	int i;

	for(i=n-1; i>=0; i--)
		unput(p, s[i]);
}

static
Rune *
addtok(HtmlParser *p, Rune *r, int val)
{
	if(r == nil)
		return nil;

	if(r == &p->tok[NTOK-1]){
		*r = L'\0';
		dprint('T', p, "token buffer too short: %S\n", p->tok);
		return nil;
	}
	*r++=val;

	return r;
}

static
void
lower(Rune *r)
{
	for(; *r; r++)
		*r = tolowerrune(*r);
}

static
Rune *
word(Rune *r)
{
	while(L'a'<=*r && *r<=L'z' || L'A'<=*r && *r<=L'Z' || L'0'<=*r && *r<=L'9')
		r++;

	return r;
}

static
Rune *
white(Rune *r)
{
	while(isspace(*r))
		r++;

	return r;
}

static
Rune*
quote(Rune *r)
{
	Rune q;

	q = *r++;
	while(*r!=q && *r != L'\0')
		r++;

	return r;
}

static
void
dnl(Rune *r)
{
	Rune *t;

	for(t=r; *r; r++) 
		if(*r!=L'\r' && *r!=L'\n')
			*t++=*r;

	*t=L'\0';
}

static
void
dsp(Rune *r)
{
	Rune *t;

	t = r;
	r = white(t);
	while(*r){
		if(isspace(*r)){
			r = white(r);
			*--r = L' ';
		}
		*t++ = *r++;
	}
	if(isspace(t[-1]))
		t--;
	*t=L'\0';
}

static
Rune *
entchar(Rune *r)
{
	while(*r=='#' || 'a'<=*r && *r<='z' || 'A'<=*r && *r<='Z' || '0'<=*r && *r<='9')
		r++;

	return r;
}

static
void 
rmentities(HtmlParser *p, Rune *s)
{
	Rune *t, *s1;
	char buf[10];
	int i, c, sep;

	t = s;
	do{
		c = *s++;
		if(c==L'&'
		&& ((*s==L'#' && (L'A'<=s[1] && s[1]<=L'Z' || L'0'<=s[1] && s[1]<=L'9'))
		  || L'a'<=*s && *s<=L'z'
		  || L'A'<=*s && *s<=L'Z')){
			s1 = entchar(s);
			sep = *s1;
			if(sep==';' || sep=='\n'){
				*s1++ = '\0';
				if(*s==L'#' && L'0'<=s[1] && s[1]<=L'9'){
					snprint(buf, sizeof(buf), "%S", s+1);
					c = atoi(buf);
					if(c == 0xa0)
						c = ' ';
					if(sep != ';')
						*--s1 = sep;
				}else{
					i = binsearch(s, htmlentities, nelem(htmlentities));
					if(i >= 0){
						c = htmlentities[i].flag1;
						if(sep != ';')
							*--s1 = sep;
					}else{
						dprint('T', p, "unknown entity %S", s);
						s1[-1] = sep;
						s1 = s;
					}
				}
				s = s1;
			}
		}
		*t++ = c;
	}while(c);
}

static
int
gettag(HtmlParser *p)
{
	Attr *a;
	Rune rune, *r, *t, *name;
	int c;

	r = p->tok;
	for(;;){
		c = input(p);
		if(c==EOF || c==EOB || c=='>')
			break;
		r = addtok(p, r, c);
	}
	if(c == EOB){
		unputstr(p, p->tok, r-p->tok);
		unput(p, '<');
		return EOB;
	}
	if(c == EOF)
		dprint('T', p, "EOF in tag");

	*r = L'\0';
	r = p->tok;
	p->tagstart = 1;
	if(*r == L'/'){
		p->tagstart = 0;
		r++;
	}
	name = r;
	p->nattr = 0;
	a = p->attr;
	r = word(r);
	if(!isspacerune(*r) && *r!=L'\0'){
		dprint('T', p, "bad tag name in %S", name);
		return NOTAG;
	}
	if(*r != L'\0')
		*r++ = L'\0';

	lower(name);
	p->tag = binsearch(name, htmltags, nelem(htmltags));
	if(p->tag < 0){
		dprint('T', p, "unknown tag: %S", name);
		return NOTAG;
	}
	for(;;){
		r = white(r);
		if(*r==L'/' || *r==L'\0'){
			a->name = nil;
			break;
		}
		a->name = r;
		r = word(r);
		if(*r==L'-' || *r==L':')
			r =  word(r+1);

		t = white(r);
		rune = *t;
		*r = L'\0';
		lower(a->name);
		if(rune == L'='){
			r = white(t+1);
			if(*r==L'\'' || *r==L'"'){
				a->value = r+1;
				r = quote(r);
				if(*r == L'\0'){
					dprint('T', p, "no terminating quote in rhs of attribute %S", a->name);
					a->name = nil;
					break;
				}
				*r++ = L'\0';
				dnl(a->value);
			}else{
				a->value=r;
				while(!isspacerune(*r) && *r!=L'\0')
					r++;
				if(*r != L'\0')
					*r++ = L'\0';
			}
			rmentities(p, a->value);
		}else{
			if(rune != L'\0')
				r++;
			a->value = L"";
		}
		if(a == &p->attr[NATTR-1]){
			dprint('T', p, "too many attributes!");
		}else{
			a++;
			p->nattr++;
		}
	}
	return TAG;
}

static
int
getspecial(HtmlParser *p, int start)
{
	int c, a[2], i;
	Rune *r;

	r = addtok(p, p->tok, start);
	for(i=0; i<2; i++){
		a[i] = input(p);
		r = addtok(p, r, a[i]);
	}
	/* <!-- comment --> */
	if(start=='!' && a[0]=='-' && a[1]=='-'){
		i = 0;
		for(;;){
			c = input(p);
			if(c == EOB){
	eob:
				unputstr(p, p->tok, r-p->tok);
				unput(p, '<');
				return EOB;
			}
			if(c==EOF || c==L'>' && i>=2)
				break;
			if(c == L'-')
				i++;
			else
				i = 0;
			r = addtok(p, r, c);
		}
	}else{
		for(;;){
			c = input(p);
			if(c == EOB)
				goto eob;
			if(c==EOF ||c=='>')
				break;
			r = addtok(p, r, c);
		}
	}
	if(c == EOF)
		dprint('T', p, "EOF in special");
	p->tok[0] = L'\0';
	return NOTAG;
}

static
int
gettext(HtmlParser *p)
{
	Rune *r;
	int c;

	r = p->tok;
	for(;;){
		c = input(p);
		if(c == EOB){
			unputstr(p, p->tok, r-p->tok);
			return EOB;
		}
		if(c == EOF)
			break;
		if(c == '<'){
			unput(p, '<');
			break;
		}
		r = addtok(p, r, c);
	}
	*r = L'\0';
	rmentities(p, p->tok);
	if(!p->pre){
		dsp(p->tok);
		dnl(p->tok);
	}
	return TEXT;
}

static
int
getscript(HtmlParser *p)
{
	Rune *r, buf[10];
	int c, c1, i, len;

	r = p->tok;
	for(;;){
		c = input(p);
		if(c == EOF){
	eof:
			dprint('T', p, "EOF in script");
			break;
		}
		if(c == EOB){
	eob:
			unputstr(p, p->tok, r-p->tok);
			return EOB;
		}
		/* "..." or '...' */
		if(c=='\'' || c=='"'){
			c1 = c;
			for(;;){
				r = addtok(p, r, c1);
				c1 = input(p);
				if(c1 == EOF)
					goto eof;
				if(c1 == EOB)
					goto eob;
				if(c1==c)
					break;	
			}
		}else if(c=='<'){
			c = input(p);
			if(c == EOF)
				goto eof;
			if(c == EOB)
				goto eob;
			if(c == '/'){
				len = runestrlen(htmltags[p->current->tag].name);
				for(i=0; i<len; i++){
					c1= input(p);
					if(c1 == EOF)
						goto eof;
					if(c1 == EOB)
						goto eob;
					buf[i] = c1;
				}
				if(runecistrncmp(buf, htmltags[p->current->tag].name, len) == 0){
					unputstr(p, buf, len);
					unputstr(p, L"</", 2);
					break;
				}else{
					for(i=0; i<len; i++)
						r = addtok(p, r, buf[i]);
				}
			}
			r = addtok(p, r, '<');
		}
		r = addtok(p, r, c);
	}
	addtok(p, r, '\0');
	return TEXT;
}

static
int
gettok(HtmlParser *p)
{
	int c;

loop:
	c = input(p);
	if(!p->pre && isspace(c))
		goto loop;

	if(c==EOF || c==EOB)
		return c;

	if(c == L'<'){
		c = input(p);
		if(c==EOB || c==EOF){
			unput(p, '<');
			return c;
		}
		 if(c==L'!' || c==L'?' || c==L'%' )
			if(c==L'!' && p->script){	/* old browser's hack */
				unputstr(p, L"<!", 2);
				return getscript(p);
			}else
				return getspecial(p, c);
		else{
			unput(p, c);
			return gettag(p);
		}
	}
	unput(p, c);
	if(p->script)
		return getscript(p);

	return gettext(p);
}

static
int
isheadel(int e)
{
	return e==Tbase || e==Tlink || e==Tmeta || e==Tnoscript || e==Tscript || e==Tstyle || e==Ttitle;
}

static
Node *
newtag(HtmlParser *p)
{
	Node *n, *parent;

	parent = p->current;
	switch(p->current->tag){

	case Thead:
		if(!isheadel(p->tag))
			parent =  p->current = p->body;
		break;

	case Ta:
		if(!htmltags[p->tag].flag2 ||  p->tag==Ta)
			parent = p->current->parent;
		break;

	case Tdd:
	case Tdt:
	case Tli:
		if(p->tag==Tdd || p->tag==Tdt || p->tag==Tli)
			parent = p->current->parent;
		break;

	case Tol:
	case Tul:
		if(p->tag != Tli)
			parent = p->current->parent;
		break;

	case Toption:
		parent = p->current->parent;
		break;

	case Tp:
	case Tplaintext:
		if(!htmltags[p->tag].flag2 && p->tag!=Ttable)
			parent = p->current->parent;
		break;

	case Ttd:
	case Tth:
		if(p->tag==Ttd || p->tag==Tth || p->tag==Ttr)
			parent = p->current->parent;
		break;

	case Ttr:
		if(p->tag == Ttr)
			parent = p->current->parent;
		break;
	}
	n = allocnode(Nelement, p->tag);
	addnode(parent, n);
	if(!htmltags[p->tag].flag1)
		p->current = n;

	return n;
}

static
void
dotag(HtmlParser *p)
{
	Node *n;
	Attr *a;
	int i;

	if(p->tagstart){
		switch(p->tag){

		case Thtml:
			n = p->current = p->html;
			break;

		case Thead:
			n = p->current = p->head;
			break;

		case Tbody:
			n = p->current =  p->body;
			break;

		default:
			n = newtag(p);
		}
		if(p->nattr){
			n->nattr = p->nattr;
			n->attr = emalloc(n->nattr*sizeof(Attr *));
			for(i=0; i<p->nattr; i++){
				a = n->attr[i] = emalloc(sizeof(Attr));
				a->name = runestrdup(p->attr[i].name);
				a->value = runestrdup(p->attr[i].value);
			}
		}
		switch(p->tag){

		case Tpre:
		case Tsamp:
			p->pre = 1;
			break;

		case Tscript:
		case Tstyle:
			p->script = 1;
			break;
		}
	}else{
		switch(p->tag){

		case Tpre:
		case Tsamp:
			p->pre = 0;
			break;

		case Tscript:
		case Tstyle:
			p->script = 0;
			break;
		}
		if(p->current->tag == p->tag)
			n = p->current->parent;
		else{
			for(n=p->current; n; n=n->parent)
				if(n->type==Nelement && n->tag==p->tag)
					break;
		}
		if(n==nil && p->tag!=Thtml)
			dprint('P', p, "endtag </%S> without starttag: ignored", htmltags[p->tag].name);
		else
			p->current = n;
	}
}

static
void
dotext(HtmlParser *p)
{
	Node *n;

	if(p->tok[0] == L'\0')
		return;

	/* document has no <head></head> */
	if(p->current->tag == Thead)
		p->current = p->body;

		n = allocnode(Ntext, -1);
		addnode(p->current, n);
		n->text = runestrdup(p->tok);
}

void
htmlparse(HtmlParser *p, char *s, int n)
{
	p->in = s;
	p->nin = n;
	if(n == 0)
		p->eof = 1;
	for(;;){
		switch(gettok(p)){

		case TAG:
			dotag(p);
			break;

		case TEXT:
			dotext(p);
			break;

		case EOF:
		case EOB:
			return;
		}
	}
}

static
int
Nconv(Fmt *fp)
{
	Node *n;
	char str[1024];
	int i, j;

	n = va_arg(fp->args, Node*);
	if(n == nil)
		return fmtprint(fp, "<nil>");

	switch(n->type){

	case Nelement:
		j = snprint(str, sizeof(str)-1, "Element: %S", htmltags[n->tag].name);
		for(i=0; i<n->nattr; i++)
			j += snprint(str+j, sizeof(str)-j-1, " \"%S=%S\"", n->attr[i]->name, n->attr[i]->value);
		break;

	case Ntext:
		snprint(str, sizeof(str)-1, "Text: \"%S\"", n->text);
		break;
	}
	return fmtstrcpy(fp, str);
}

void
htmlinit(HtmlParser *p, Rune *name)
{
	if(!installed){
		fmtinstall('N', Nconv);
		installed++;
	}
	p->yysptr = p->yysbuf;
	p->name = name;
	p->lineno = 1;
	p->eof = p->script = p->pre = 0;
	p->html  = allocnode(Nelement, Thtml);
	p->head = allocnode(Nelement, Thead);
	addnode(p->html, p->head);
	p->body = allocnode(Nelement, Tbody);
	addnode(p->html, p->body);
	p->current = p->head;
}

static
void
prtext(Node *n)
{
	Rune *s;
	int nl;

	if(n->type == Ntext){
		print("%S ", n->text);
		return;
	}

	nl = 0;
	switch(n->tag){

	case Tscript:
	case Tstyle:
		return;

	case Thr:
		print("\n=======\n");
		return;

	case Tli:
		print("• ");
		/* fall through */
	case Taddress:
	case Tblockquote:
	case Tbr:
	case Tdd:
	case Th1:
	case Th2:
	case Th3:
	case Th4:
	case Th5:
	case Th6:
	case Ttr:
	case Ttd:
	case Tul:
		nl++;
		break;

	case Timg:
		s = getattr(n, L"alt");
		if(s && runestrlen(s))
			print("[%S] ", s);
		break;

	}
	if(n->child)
		for(n=n->child[0]; n; n=n->next)
			prtext(n);
	if(nl)
		print("\n");
}

void
htmlprtext(Node *n)
{
	prtext(n);
	print("\n");
}

static
void
prtree(Node *n, int d)
{
	int i;

 	if(n == nil)
		return;

	for(i=0; i<d; i++)
		print("   ");

	print("%N", n);
	if(n->child){
		print(" {\n");
		for(i=0; i<n->nchild; i++)
			prtree(n->child[i], d+1);
		for(i=0; i<d; i++)
			print("   ");
		print("}");
	}
	print("\n");
}

void
htmlprtree(Node *n)
{
	prtree(n, 0);
	print("\n");
}
